from __future__ import print_function
import numpy as np
import os, sys
import random
import argparse
import datetime
import tensorflow as tf
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.datasets import cifar10
import tensorflow.keras.applications as applications

def parse_command_line():
    parser = argparse.ArgumentParser()
    parser.add_argument("--device", required=False, type=str, choices=["cpu", "cuda"], default="cuda")
    parser.add_argument("--strategy", required=False, type=str, choices=["mirrored", "multi-worker"], default="multi-worker")
    parser.add_argument("--num_epochs", required=False, type=int, default=5)
    parser.add_argument("--batch_size", required=False, type=int, default=128)
    parser.add_argument("--tensorboard", required=False, help="Whether to use tensorboard callback", action="store_true", default=False)
    parser.add_argument("--profile_batches", required=False, help='Batches to profile with for tensorboard. Format "batch_start,batch_end"', type=str, default="2,5")
    args = parser.parse_args()

    # default args for distributed
    args.world_size = int(os.environ["WORLD_SIZE"])
    args.world_rank = int(os.environ["RANK"])
    args.local_rank = int(os.environ["LOCAL_RANK"])
    args.global_batch_size = args.batch_size * args.world_size
    args.verbosity = 2 if args.world_rank == 0 else 0 # only use verbose for master process

    if args.world_rank == 0:
        print("Settings:")
        settings_map = vars(args)
        for name in sorted(settings_map.keys()):
            print("--" + str(name) + ": " + str(settings_map[name]))
        print("")

    return args

def preprocess_data(images, labels):
    images = tf.image.resize(images, (224, 224))  # Resize for ResNet-50
    images = images / 255.0  # Normalize to [0, 1]
    return images, labels

def load_dataset(args):
    # load the cifar10 data and generate input pipelines
    (x_train, y_train), (x_test, y_test) = cifar10.load_data()

    # convert class vectors to binary class matrices.
    y_train = tf.keras.utils.to_categorical(y_train, 10)
    y_test = tf.keras.utils.to_categorical(y_test, 10)

    ds_train = (
        tf.data.Dataset.from_tensor_slices((x_train, y_train))
        .map(preprocess_data)
        .shuffle(x_train.shape[0])
        .cache()
        .batch(args.global_batch_size)
        .prefetch(tf.data.experimental.AUTOTUNE)
    )

    ds_test = (
        tf.data.Dataset.from_tensor_slices((x_test, y_test))
        .map(preprocess_data)
        .cache()
        .batch(args.global_batch_size)
        .prefetch(tf.data.experimental.AUTOTUNE)
    )

    # get updated shapes
    train_shape, test_shape = ds_train.element_spec[0].shape, ds_test.element_spec[0].shape
    if args.world_rank == 0:
        print(f"train_shape:", x_train.shape, " -> ", train_shape)
        print(f"test_shape:", x_test.shape, " -> ", test_shape)

    return ds_train, ds_test, train_shape

def setup(args):
    print(f"Number of GPU devices found on worker {args.world_rank}:", len(tf.config.list_physical_devices("GPU")))

    tf.keras.backend.clear_session()
    tf.config.optimizer.set_jit(True)

    # define data parallel strategy for distrbuted training
    if args.strategy == "mirrored":
        strategy = tf.distribute.MirroredStrategy()
    else:
        strategy = tf.distribute.MultiWorkerMirroredStrategy(
            communication_options=tf.distribute.experimental.CommunicationOptions(
                implementation=tf.distribute.experimental.CollectiveCommunication.NCCL
            )
        )
        print("MultiWorkerMirroredStrategy.num_replicas_in_sync:", strategy.num_replicas_in_sync)
        print("MultiWorkerMirroredStrategy.worker_index:", strategy.cluster_resolver.task_id)

    return strategy

def main():
    # always use the same seed
    random.seed(42)
    tf.random.set_seed(42)
    np.random.seed(42)

    # parse command line arguments
    args = parse_command_line()

    # run setup (e.g., create distributed environment if desired)
    strategy = setup(args)

    # loading desired dataset
    ds_train, ds_test, train_shape = load_dataset(args)

    # callbacks to register
    callbacks = []

    # create and compile the model
    with strategy.scope():
        model = applications.ResNet50(weights=None, input_shape=train_shape[1:], classes=10)
        cur_optimizer = Adam(learning_rate=0.001 * args.world_size) # scale learning rate with number of workers
        model.compile(loss="categorical_crossentropy", optimizer=cur_optimizer, metrics=["accuracy"])

    # callbacks to register
    if args.tensorboard:
        tensorboard_callback = tf.keras.callbacks.TensorBoard(
            log_dir=os.path.join("logs", datetime.datetime.now().strftime("%Y%m%d-%H%M%S")),
            histogram_freq=1,
            profile_batch=args.profile_batches,
        )
        callbacks.append(tensorboard_callback)

    # train the model
    model.fit(ds_train, epochs=args.num_epochs, verbose=args.verbosity, callbacks=callbacks)

    # evaluate model
    scores = model.evaluate(ds_test, verbose=args.verbosity)
    if args.world_rank == 0:
        print(f"Test Evaluation: Accuracy: {scores[1]}")

if __name__ == "__main__":
    main()
